﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.IO;

namespace FTPLibrary
{
    public class FTPManager : IDisposable
    {
        #region 定数

        /// <summary>
        /// ディレクトリパーミッション識別子
        /// </summary>
        private const string ID_DIRECTORY = "d";
        /// <summary>
        /// ファイルパーミッション識別子
        /// </summary>
        private const string ID_FILE = "-";

        #endregion

        #region 列挙体

        /// <summary>
        /// リストタイプ
        /// </summary>
        public enum ListType : int
        {
            None = -1,
            Folder = 0,
            File = 1,
            RootDir = 2
        }

        #endregion

        #region リスト情報クラス
        /// <summary>
        /// リスト情報クラス
        /// </summary>
        public class TargetInfo
        {

            private ListType _type;
            private string _name;

            public ListType TypeInfo
            {
                set
                {
                    this._type = value;
                }
                get
                {
                    return this._type;
                }
            }

            public string Name
            {
                set
                {
                    this._name = value;
                }
                get
                {
                    return this._name;
                }
            }

            public TargetInfo()
            {
                this._type = ListType.None;
                this._name = "";
            }

            public TargetInfo(string name, ListType info)
                : this()
            {
                this._type = info;
                this._name = name;
            }
        }
        #endregion

        #region メンバ変数

        /// <summary>
        /// ユーザー名
        /// </summary>
        private string _userName;

        /// <summary>
        /// パスワード
        /// </summary>
        private string _password;

        /// <summary>
        /// サーバー名
        /// </summary>
        private string _ftpserverName;

        /// <summary>
        /// カレントディレクトリ
        /// </summary>
        private string _currentDirectory;

        /// <summary>
        /// カレントディレクトリ情報リスト
        /// </summary>
        private List<TargetInfo> _currentFilesList;

        /// <summary>
        /// 接続可能状態
        /// </summary>
        private bool _connectable;

        /// <summary>
        /// 接続状態
        /// </summary>
        private bool _connected;

        /// <summary>
        /// 接続情報
        /// </summary>
        private NetworkCredential _ftpCredential;

        private Encoding _ec;
        #endregion

        #region プロパティ
        /// <summary>
        /// ユーザー名
        /// </summary>
        public string UserName
        {
            set
            {
                this._userName = value;
            }
            get
            {
                return this._userName;
            }
        }

        /// <summary>
        /// パスワード
        /// </summary>
        public string Password
        {
            set
            {
                this._password = value;
            }
            get
            {
                return this._password;
            }
        }

        /// <summary>
        /// カレントディレクトリ
        /// </summary>
        public string CurrentDirectory
        {
            get
            {
                return this._ftpserverName + this._currentDirectory;
            }
        }

        /// <summary>
        /// カレントディレクトリ情報
        /// </summary>
        public List<TargetInfo> DirectoryInfo
        {
            get
            {
                return this._currentFilesList;
            }
        }
        #endregion

        #region コンストラクタ

        /// <summary>
        /// コンストラクタ
        /// </summary>
        private FTPManager()
        {
            this._connected = false;
            this._connectable = false;

            this._userName = "";
            this._password = "";
            this._ftpserverName = "ftp://";
            this._currentDirectory = "";

            this._ec = Encoding.GetEncoding("Shift-Jis");
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="ftpServerName"></param>
        public FTPManager(string ftpServerName)
            : this()
        {
            this._ftpserverName = this._ftpserverName + ftpServerName;
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="ftpServerName"></param>
        /// <param name="userName"></param>
        /// <param name="password"></param>
        public FTPManager(string ftpServerName, string userName, string password)
            : this(ftpServerName)
        {
            this.SetConnect(userName, password);

            GetCurrentInfo(this._ftpserverName);
        }

        #endregion

        #region デストラクタ
        /// <summary>
        /// デストラクタ
        /// </summary>
        ~FTPManager()
        {
            this.Dispose();
        }
        #endregion

        #region Disposeメソッド
        /// <summary>
        /// Disposeメソッド
        /// </summary>
        public void Dispose()
        {
            if (this._connected)
            {
                this.Disconnect();
            }
        }
        #endregion

        #region 接続情報格納処理
        /// <summary>
        /// 接続情報格納
        /// </summary>
        private void SetConnect()
        {
            this._ftpCredential = new System.Net.NetworkCredential(this._userName, this.Password);
            this._connectable = true;
        }

        /// <summary>
        /// 接続情報格納処理
        /// </summary>
        /// <param name="userName"></param>
        /// <param name="password"></param>
        public void SetConnect(string userName, string password)
        {
            this._userName = userName;
            this._password = password;

            SetConnect();
        }
        #endregion

        #region 切断処理
        /// <summary>
        /// 切断処理
        /// </summary>
        private void Disconnect()
        {
            if (this._connectable)
            {
                FtpWebRequest ftpReq = (FtpWebRequest)WebRequest.Create(this._ftpserverName);
                if (this._ftpCredential != null)
                {
                    ftpReq.Credentials = this._ftpCredential;
                }
                ftpReq.Method = WebRequestMethods.Ftp.PrintWorkingDirectory;
                ftpReq.KeepAlive = false;

                using (FtpWebResponse ftpRes = (FtpWebResponse)ftpReq.GetResponse())
                {
                    //切断
                    ftpRes.Close();
                }
            }
        }
        #endregion

        #region カレントディレクトリ情報格納処理
        /// <summary>
        /// カレントディレクトリ情報格納処理
        /// </summary>
        private void GetCurrentInfo(string path)
        {
            List<string> getDatas;

            if (this._connectable)
            {
                FtpWebRequest ftpReq = (FtpWebRequest)WebRequest.Create(path);
                if (this._ftpCredential != null)
                {
                    ftpReq.Credentials = this._ftpCredential;
                }
                ftpReq.Method = WebRequestMethods.Ftp.PrintWorkingDirectory;

                this._currentFilesList = new List<TargetInfo>();

                using (FtpWebResponse ftpRes = (FtpWebResponse)ftpReq.GetResponse())
                {
                    //カレントディレクトリ格納
                    getDatas = new List<string>(ftpRes.StatusDescription.Split('"'));
                    this._currentDirectory = getDatas[1];
                    //切断
                    ftpRes.Close();
                    this._connected = true;

                    GetDirectoryInfo();
                }
            }
        }
        #endregion

        #region ディレクトリ情報取得処理
        /// <summary>
        /// ディレクトリ情報取得処理
        /// </summary>
        private void GetDirectoryInfo()
        {
            List<string> getDataDetails;
            List<string> detailData;
            string readData;

            if (this._connectable)
            {
                FtpWebRequest ftpReq = (FtpWebRequest)WebRequest.Create(this._ftpserverName + "/%2f" + this._currentDirectory + "/");
                if (this._ftpCredential != null)
                {
                    ftpReq.Credentials = this._ftpCredential;
                }
                ftpReq.Method = WebRequestMethods.Ftp.ListDirectoryDetails;
                ftpReq.UsePassive = false;

                using (FtpWebResponse ftpRes = (FtpWebResponse)ftpReq.GetResponse())
                {
                    getDataDetails = new List<string>();
                    using (StreamReader sr = new StreamReader(ftpRes.GetResponseStream()))
                    {
                        //ストリームより取得データを格納
                        readData = sr.ReadToEnd();

                        //カレンドディレクトリ
                        getDataDetails = new List<string>(readData.Split(new string[1] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries));
                    }

                    //切断
                    ftpRes.Close();
                    this._connected = true;

                    //ディレクトリ情報格納

                    //ルートディレクトリ追加
                    this._currentFilesList.Add(new TargetInfo("..", ListType.RootDir));

                    foreach (string data in getDataDetails)
                    {
                        TargetInfo ti = new TargetInfo();
                        string id = data.Substring(0, 1);
                        detailData = new List<string>(data.Split(' '));

                        switch (id)
                        {
                            case ID_DIRECTORY:
                                //ディレクトリ
                                ti.TypeInfo = ListType.Folder;
                                break;
                            case ID_FILE:
                                //ディレクトリ
                                ti.TypeInfo = ListType.File;
                                break;
                            default:
                                break;
                        }
                        ti.Name = detailData[detailData.Count - 1];

                        if (ti.TypeInfo != ListType.None)
                        {
                            this._currentFilesList.Add(ti);
                        }
                    }
                }
            }
        }
        #endregion

        #region ディレクトリ移動処理
        /// <summary>
        /// ディレクトリ移動処理
        /// </summary>
        /// <param name="dirName"></param>
        public void ChangeDirectory(string dirName)
        {
            GetCurrentInfo(this._ftpserverName + "/%2f" + this._currentDirectory + "/" + dirName);
        }
        #endregion

        #region ファイルダウンロード処理
        /// <summary>
        /// ファイルダウンロード処理
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="savePath"></param>
        /// <param name="readBinary"></param>
        /// <returns></returns>
        private int DownLoadFile(string fileName, string savePath, bool readBinary)
        {
            int ret = -1;
            int readSize;
            byte[] readBuffer = new byte[1024];

            if (this._connectable)
            {
                FtpWebRequest ftpReq = (FtpWebRequest)WebRequest.Create(this._ftpserverName + "/%2f" + this._currentDirectory + "/" + fileName);
                if (this._ftpCredential != null)
                {
                    ftpReq.Credentials = this._ftpCredential;
                }
                ftpReq.Method = WebRequestMethods.Ftp.DownloadFile;
                ftpReq.UseBinary = readBinary;
                ftpReq.UsePassive = false;

                using (FtpWebResponse ftpRes = (FtpWebResponse)ftpReq.GetResponse())
                {
                    using (Stream resStream = ftpRes.GetResponseStream())
                    {
                        using (FileStream fs = new FileStream(savePath, FileMode.Create, FileAccess.Write))
                        {
                            using (StreamWriter sw = new StreamWriter(fs))
                            {
                                while (true)
                                {
                                    readSize = resStream.Read(readBuffer, 0, readBuffer.Length);
                                    if (readSize == 0)
                                    {
                                        break;
                                    }
                                    sw.Write(this._ec.GetString(readBuffer, 0, readSize));
                                }

                                sw.Close();
                            }

                            fs.Close();

                            ret = 0;
                        }

                        resStream.Close();
                    }

                    ftpRes.Close();
                }
            }

            return ret;
        }

        /// <summary>
        /// ファイルダウンロード処理（バイナリ転送）
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="savePath"></param>
        /// <returns></returns>
        public int DownLoadFileBinary(string fileName, string savePath)
        {
            return DownLoadFile(fileName, savePath, true);
        }

        /// <summary>
        /// ファイルダウンロード処理（ASCII転送）
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="savePath"></param>
        /// <returns></returns>
        public int DownLoadFileAscii(string fileName, string savePath)
        {
            return DownLoadFile(fileName, savePath, false);
        }
        #endregion

        #region ファイルアップロード処理
        /// <summary>
        /// ファイルアップロード処理
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="savePath"></param>
        /// <param name="readBinary"></param>
        /// <returns></returns>
        private int UploadFile(string fileName, string savePath, bool readBinary)
        {
            int ret = -1;
            string readData;
            byte[] readBuffer;

            if (this._connectable)
            {
                FtpWebRequest ftpReq = (FtpWebRequest)WebRequest.Create(this._ftpserverName + "/%2f" + this._currentDirectory + "/" + fileName);
                if (this._ftpCredential != null)
                {
                    ftpReq.Credentials = this._ftpCredential;
                }
                ftpReq.Method = WebRequestMethods.Ftp.UploadFile;
                ftpReq.UseBinary = readBinary;
                ftpReq.UsePassive = false;

                using (FileStream fs = new FileStream(savePath, FileMode.Open, FileAccess.Read))
                {
                    using (StreamReader sr = new StreamReader(fs))
                    {
                        using (Stream reqStream = ftpReq.GetRequestStream())
                        {
                            //ファイル終端迄読込
                            readData = sr.ReadToEnd();

                            //文字コードを指定してバイト配列に変換
                            readBuffer = this._ec.GetBytes(readData);


                            //アップロード
                            reqStream.Write(readBuffer, 0, readBuffer.Length);

                            reqStream.Close();

                            ret = 0;
                        }

                        sr.Close();
                    }
                    fs.Close();
                }

                //ステータス受信
                using (FtpWebResponse ftpRes = (FtpWebResponse)ftpReq.GetResponse())
                {
                    ftpRes.Close();
                }
            }

            return ret;
        }

        /// <summary>
        /// ファイルアップロード処理（バイナリ転送）
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="savePath"></param>
        /// <returns></returns>
        public int UpLoadFileBinary(string fileName, string savePath)
        {
            return UploadFile(fileName, savePath, true);
        }

        /// <summary>
        /// ファイルアップロード処理（ASCII転送）
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="savePath"></param>
        /// <returns></returns>
        public int UpLoadFileAscii(string fileName, string savePath)
        {
            return UploadFile(fileName, savePath, false);
        }
        #endregion

        #region ファイル削除処理
        /// <summary>
        /// ファイル削除処理
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public int DeleteFile(string fileName)
        {
            int ret = -1;

            if (this._connectable)
            {
                FtpWebRequest ftpReq = (FtpWebRequest)WebRequest.Create(this._ftpserverName + "/%2f" + this._currentDirectory + "/" + fileName);
                if (this._ftpCredential != null)
                {
                    ftpReq.Credentials = this._ftpCredential;
                }
                ftpReq.Method = WebRequestMethods.Ftp.DeleteFile;

                using (FtpWebResponse ftpRes = (FtpWebResponse)ftpReq.GetResponse())
                {
                    //切断
                    ftpRes.Close();
                    this._connected = true;
                }
            }

            return ret;
        }
        #endregion
    }
}
